"""
Utils for the lyra package.
"""
import logging
import sys

from rich.logging import RichHandler


def get_logger():
    """Get the logger."""
    logger = logging.getLogger(__name__)
    formatter = logging.Formatter("%(message)s")

    # we check if the logger already has a handler
    # to avoid adding multiple handlers
    if logger.hasHandlers():
        return logger
    if sys.stdout.isatty():
        handler = RichHandler(
            markup=False,
            rich_tracebacks=True,
            locals_max_string=None,
            locals_max_length=None,
        )
    else:
        handler = logging.StreamHandler(sys.stdout)

    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)
    return logger



"""
Async client for Lyra
"""

import asyncio
import json
import time
from datetime import datetime

import aiohttp
from web3 import Web3

from lyra.constants import CONTRACTS, TEST_PRIVATE_KEY
from lyra.enums import Environment, InstrumentType, OrderSide, OrderType, TimeInForce, UnderlyingCurrency
from lyra.utils import get_logger
from lyra.ws_client import WsClient as BaseClient


class AsyncClient(BaseClient):
    """
    We use the async client to make async requests to the lyra API
    We us the ws client to make async requests to the lyra ws API
    """

    current_subscriptions = {}

    listener = None
    subscribing = False
    _ws = None

    def __init__(
        self,
        private_key: str = TEST_PRIVATE_KEY,
        env: Environment = Environment.TEST,
        logger=None,
        verbose=False,
        subaccount_id=None,
        wallet=None,
    ):
        """
        Initialize the LyraClient class.
        """
        self.verbose = verbose
        self.env = env
        self.contracts = CONTRACTS[env]
        self.logger = logger or get_logger()
        self.web3_client = Web3()
        self.signer = self.web3_client.eth.account.from_key(private_key)
        self.wallet = self.signer.address if not wallet else wallet
        print(f"Signing address: {self.signer.address}")
        if wallet:
            print(f"Using wallet: {wallet}")
        self.subaccount_id = subaccount_id
        print(f"Using subaccount id: {self.subaccount_id}")
        self.message_queues = {}
        self.connecting = False
        # we make sure to get the event loop

    @property
    async def ws(self):
        if self._ws is None:
            self._ws = await self.connect_ws()
        if not self._ws.connected:
            self._ws = await self.connect_ws()
        return self._ws

    async def fetch_ticker(self, instrument_name: str):
        """
        Fetch the ticker for a symbol
        """
        id = str(int(time.time()))
        payload = {"instrument_name": instrument_name}
        if not self.ws:
            await self.connect_ws()

        self.ws.send(json.dumps({"method": "public/get_ticker", "params": payload, "id": id}))

        # we now wait for the response
        while True:
            response = self.ws.recv()
            response = json.loads(response)
            if response["id"] == id:
                close = float(response["result"]["best_bid_price"]) + float(response["result"]["best_ask_price"]) / 2
                response["result"]["close"] = close
                return response["result"]

    def get_subscription_id(self, instrument_name: str, group: str = "1", depth: str = "100"):
        return f"orderbook.{instrument_name}.{group}.{depth}"

    async def subscribe(self, instrument_name: str, group: str = "1", depth: str = "100"):
        """
        Subscribe to the order book for a symbol
        """
        # if self.listener is None or self.listener.done():
        asyncio.create_task(self.listen_for_messages())
        channel = self.get_subscription_id(instrument_name, group, depth)
        if channel not in self.message_queues:
            self.message_queues[channel] = asyncio.Queue()
            msg = {"method": "subscribe", "params": {"channels": [channel]}}
            await self.ws.send_json(msg)
            return

        while instrument_name not in self.current_subscriptions:
            await asyncio.sleep(0.01)
        return self.current_subscriptions[instrument_name]

    async def connect_ws(self):
        self.connecting = True
        session = aiohttp.ClientSession()
        ws = await session.ws_connect(self.contracts['WS_ADDRESS'])
        self._ws = ws
        self.connecting = False
        return ws

    async def listen_for_messages(
        self,
    ):
        while True:
            try:
                msg = await self.ws.receive_json()
            except TypeError:
                continue
            if "error" in msg:
                print(msg)
                raise Exception(msg["error"])
            if "result" in msg:
                result = msg["result"]
                if "status" in result:
                    # print(f"Succesfully subscribed to {result['status']}")
                    for channel, value in result['status'].items():
                        # print(f"Channel {channel} has value {value}")
                        if "error" in value:
                            raise Exception(f"Subscription error for channel: {channel} error: {value['error']}")
                    continue
            #  default to putting the message in the queue
            subscription = msg['params']['channel']
            data = msg['params']['data']
            self.handle_message(subscription, data)

    async def login_client(
        self,
    ):
        login_request = {
            'method': 'public/login',
            'params': self.sign_authentication_header(),
            'id': str(int(time.time())),
        }
        await self._ws.send_json(login_request)
        async for data in self._ws:
            message = json.loads(data.data)
            if message['id'] == login_request['id']:
                if "result" not in message:
                    raise Exception(f"Unable to login {message}")
                break

    def handle_message(self, subscription, data):
        bids = data['bids']
        asks = data['asks']

        bids = list(map(lambda x: (float(x[0]), float(x[1])), bids))
        asks = list(map(lambda x: (float(x[0]), float(x[1])), asks))

        instrument_name = subscription.split(".")[1]

        if subscription in self.current_subscriptions:
            old_params = self.current_subscriptions[subscription]
            _asks, _bids = old_params["asks"], old_params["bids"]
            if not asks:
                asks = _asks
            if not bids:
                bids = _bids
        timestamp = data['timestamp']
        datetime_str = datetime.fromtimestamp(timestamp / 1000)
        nonce = data['publish_id']
        self.current_subscriptions[instrument_name] = {
            "asks": asks,
            "bids": bids,
            "timestamp": timestamp,
            "datetime": datetime_str.isoformat(),
            "nonce": nonce,
            "symbol": instrument_name,
        }
        return self.current_subscriptions[instrument_name]

    async def watch_order_book(self, instrument_name: str, group: str = "1", depth: str = "100"):
        """
        Watch the order book for a symbol
        orderbook.{instrument_name}.{group}.{depth}
        """

        if not self.ws and not self.connecting:
            await self.connect_ws()
            await self.login_client()

        subscription = self.get_subscription_id(instrument_name, group, depth)

        if subscription not in self.message_queues:
            while any([self.subscribing, self.ws is None, self.connecting]):
                await asyncio.sleep(1)
            await self.subscribe(instrument_name, group, depth)

        while instrument_name not in self.current_subscriptions and not self.connecting:
            await asyncio.sleep(0.01)

        return self.current_subscriptions[instrument_name]

    async def fetch_instruments(
        self,
        expired=False,
        instrument_type: InstrumentType = InstrumentType.PERP,
        currency: UnderlyingCurrency = UnderlyingCurrency.BTC,
    ):
        return super().fetch_instruments(expired, instrument_type, currency)

    async def close(self):
        """
        Close the connection
        """
        self.ws.close()

    async def fetch_tickers(
        self,
        instrument_type: InstrumentType = InstrumentType.OPTION,
        currency: UnderlyingCurrency = UnderlyingCurrency.BTC,
    ):
        if not self._ws:
            await self.connect_ws()
        instruments = await self.fetch_instruments(instrument_type=instrument_type, currency=currency)
        instrument_names = [i['instrument_name'] for i in instruments]
        id_base = str(int(time.time()))
        ids_to_instrument_names = {
            f'{id_base}_{enumerate}': instrument_name for enumerate, instrument_name in enumerate(instrument_names)
        }
        for id, instrument_name in ids_to_instrument_names.items():
            payload = {"instrument_name": instrument_name}
            await self._ws.send_json({'method': 'public/get_ticker', 'params': payload, 'id': id})
            await asyncio.sleep(0.1)  # otherwise we get rate limited...
        results = {}
        while ids_to_instrument_names:
            message = await self._ws.receive()
            if message is None:
                continue
            if 'error' in message:
                raise Exception(f"Error fetching ticker {message}")
            if message.type == aiohttp.WSMsgType.CLOSED:
                # we try to reconnect
                print(f"Erorr fetching ticker {message}...")
                self._ws = await self.connect_ws()
                return await self.fetch_tickers(instrument_type, currency)
            message = json.loads(message.data)
            if message['id'] in ids_to_instrument_names:
                try:
                    results[message['result']['instrument_name']] = message['result']
                except KeyError:
                    print(f"Error fetching ticker {message}")
                del ids_to_instrument_names[message['id']]
        return results

    async def get_collaterals(self):
        breakpoint()
        return super().get_collaterals()

    async def get_positions(self, currency: UnderlyingCurrency = UnderlyingCurrency.BTC):
        return super().get_positions()

    async def get_open_orders(self, status, currency: UnderlyingCurrency = UnderlyingCurrency.BTC):
        return super().fetch_orders(
            status=status,
        )

    async def create_order(
        self,
        price,
        amount,
        instrument_name: str,
        reduce_only=False,
        side: OrderSide = OrderSide.BUY,
        order_type: OrderType = OrderType.LIMIT,
        time_in_force: TimeInForce = TimeInForce.GTC,
    ):
        """
        Create the order.
        """
        if side.name.upper() not in OrderSide.__members__:
            raise Exception(f"Invalid side {side}")
        order = self._define_order(
            instrument_name=instrument_name,
            price=price,
            amount=amount,
            side=side,
        )
        _currency = UnderlyingCurrency[instrument_name.split("-")[0]]
        if instrument_name.split("-")[1] == "PERP":
            instruments = await self.fetch_instruments(instrument_type=InstrumentType.PERP, currency=_currency)
            instruments = {i['instrument_name']: i for i in instruments}
            base_asset_sub_id = instruments[instrument_name]['base_asset_sub_id']
            instrument_type = InstrumentType.PERP
        else:
            instruments = await self.fetch_instruments(instrument_type=InstrumentType.OPTION, currency=_currency)
            instruments = {i['instrument_name']: i for i in instruments}
            base_asset_sub_id = instruments[instrument_name]['base_asset_sub_id']
            instrument_type = InstrumentType.OPTION

        signed_order = self._sign_order(order, base_asset_sub_id, instrument_type, _currency)
        response = self.submit_order(signed_order)
        return response



"""
Cli module in order to allow interaction.
"""
import os

import pandas as pd
import rich_click as click
from dotenv import load_dotenv
from rich import print

from lyra.analyser import PortfolioAnalyser
from lyra.enums import (
    CollateralAsset,
    Environment,
    InstrumentType,
    OrderSide,
    OrderStatus,
    OrderType,
    SubaccountType,
    UnderlyingCurrency,
)
from lyra.lyra import LyraClient
from lyra.utils import get_logger

click.rich_click.USE_RICH_MARKUP = True
pd.set_option('display.precision', 2)


def set_logger(ctx, level):
    """Set the logger."""
    if not hasattr(ctx, "logger"):
        ctx.logger = get_logger()
        ctx.logger.setLevel(level)
    return ctx.logger


def set_client(ctx):
    """Set the client."""
    # we use dotenv to load the env vars from DIRECTORY where the cli tool is executed
    _path = os.getcwd()
    env_path = os.path.join(_path, ".env")
    load_dotenv(dotenv_path=env_path)
    if not hasattr(ctx, "client"):
        auth = {
            "private_key": os.environ.get("ETH_PRIVATE_KEY"),
            "logger": ctx.logger,
            "verbose": ctx.logger.level == "DEBUG",
        }
        chain = os.environ.get("ENVIRONMENT")
        if chain == Environment.PROD.value:
            env = Environment.PROD
        else:
            env = Environment.TEST

        subaccount_id = os.environ.get("SUBACCOUNT_ID", None)
        if subaccount_id:
            subaccount_id = int(subaccount_id)
        wallet = os.environ.get("WALLET")
        ctx.client = LyraClient(**auth, env=env, subaccount_id=subaccount_id, wallet=wallet)

    if ctx.logger.level == "DEBUG":
        print(f"Client created for environment `{ctx.client.env.value}`")
    return ctx.client


@click.group("Lyra Client")
@click.option(
    "--log-level",
    "-l",
    default="INFO",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
    help="Logging level.",
)
@click.pass_context
def cli(ctx, log_level):
    """Lyra v2 client command line interface."""
    ctx.ensure_object(dict)
    ctx.obj["logger"] = set_logger(ctx, log_level)
    ctx.obj["client"] = set_client(ctx)


@cli.group("instruments")
def instruments():
    """Interact with markets."""


@cli.group("tickers")
def tickers():
    """Interact with tickers."""


@cli.group("subaccounts")
def subaccounts():
    """Interact with subaccounts."""


@cli.group("orders")
def orders():
    """Interact with orders."""


@cli.group("collateral")
def collateral():
    """Interact with collateral."""


@cli.group("positions")
def positions():
    """Interact with positions."""


@cli.group("mmp")
def mmp():
    """Interact with market making parameters."""


@mmp.command("fetch")
@click.option(
    "--underlying-currency",
    "-u",
    type=click.Choice([f.value for f in UnderlyingCurrency]),
    default=UnderlyingCurrency.ETH.value,
)
@click.option(
    "--subaccount-id",
    "-s",
    type=int,
    required=True,
)
@click.pass_context
def fetch_mmp(ctx, underlying_currency, subaccount_id):
    """Fetch market making parameters."""
    client = ctx.obj["client"]
    mmp = client.get_mmp_config(subaccount_id=subaccount_id, currency=UnderlyingCurrency(underlying_currency))
    print(mmp)


@mmp.command("set")
@click.option(
    "--underlying-currency",
    "-u",
    type=click.Choice([f.value for f in UnderlyingCurrency]),
    default=UnderlyingCurrency.ETH.value,
)
@click.option(
    "--subaccount-id",
    "-s",
    type=int,
    required=True,
)
@click.option(
    "--frozen-time",
    "-f",
    type=int,
    required=True,
    help="Time interval in ms setting how long the subaccount is frozen after an mmp trigger, "
    + "if 0 then a manual reset would be required via private/reset_mmp",
    default=10000,
)
@click.option(
    "--interval",
    "-i",
    type=int,
    required=True,
    help="Time interval in ms over which the limits are monotored, if 0 then mmp is disabled",
    default=10000,
)
@click.option(
    "--amount-limit",
    "-a",
    type=float,
    help="Maximum total order amount that can be traded within the mmp_interval across all "
    + "instruments of the provided currency. The amounts are not netted, so a filled bid "
    + "of 1 and a filled ask of 2 would count as 3.",
    default=5,
)
@click.option(
    "--delta-limit",
    "-d",
    type=float,
    help="Maximum total delta that can be traded within the mmp_interval across all instruments "
    + "of the provided currency. This quantity is netted, so a filled order with +1 delta and "
    + "a filled order with -2 delta would count as 3",
    default=2,
)
@click.pass_context
def set_mmp(ctx, underlying_currency, subaccount_id, frozen_time, interval, amount_limit, delta_limit):
    """Set market making parameters."""
    client = ctx.obj["client"]
    mmp = client.set_mmp_config(
        subaccount_id=subaccount_id,
        currency=UnderlyingCurrency(underlying_currency),
        mmp_frozen_time=frozen_time,
        mmp_interval=interval,
        mmp_amount_limit=amount_limit,
        mmp_delta_limit=delta_limit,
    )
    print(mmp)


@positions.command("fetch")
@click.pass_context
def fetch_positions(ctx):
    """Fetch positions."""
    print("Fetching positions")
    client = ctx.obj["client"]
    positions = client.get_positions()
    print(positions)


@collateral.command("fetch")
@click.pass_context
def fetch_collateral(ctx):
    """Fetch collateral."""
    print("Fetching collateral")
    client = ctx.obj["client"]
    collateral = client.get_collaterals()
    print(collateral)


@instruments.command("fetch")
@click.pass_context
@click.option(
    "--instrument-type",
    "-i",
    type=click.Choice([f.value for f in InstrumentType]),
    default=InstrumentType.PERP.value,
)
@click.option(
    "--currency",
    "-c",
    type=click.Choice([f.value for f in UnderlyingCurrency]),
    default=UnderlyingCurrency.ETH.value,
)
def fetch_instruments(ctx, instrument_type, currency):
    """Fetch markets."""
    client = ctx.obj["client"]
    markets = client.fetch_instruments(
        instrument_type=InstrumentType(instrument_type), currency=UnderlyingCurrency(currency)
    )
    print(markets)


@tickers.command("fetch")
@click.pass_context
@click.argument(
    "instrument_name",
    type=str,
)
def fetch_tickers(ctx, instrument_name):
    """Fetch tickers."""
    client = ctx.obj["client"]
    ticker = client.fetch_ticker(instrument_name=instrument_name)
    print(ticker)


@collateral.command("transfer")
@click.pass_context
@click.option(
    "--amount",
    "-a",
    type=float,
    required=True,
)
@click.option(
    "--to",
    "-t",
    type=int,
    required=True,
    help="Subaccount ID to transfer to",
)
@click.option(
    "--asset",
    "-s",
    type=click.Choice([f.value for f in CollateralAsset]),
    default=CollateralAsset.USDC.value,
)
def transfer_collateral(ctx, amount, to, asset):
    """Transfer collateral."""
    client = ctx.obj["client"]
    result = client.transfer_collateral(amount=amount, to=to, asset=CollateralAsset(asset))

    print(result)


@subaccounts.command("all")
@click.pass_context
def fetch_subaccounts(ctx):
    """Fetch subaccounts."""
    print("Fetching subaccounts")
    client = ctx.obj["client"]
    subaccounts = client.fetch_subaccounts()
    print(subaccounts)


@subaccounts.command("fetch")
@click.argument(
    "subaccount_id",
    type=int,
)
@click.option(
    "--underlying-currency",
    "-u",
    type=click.Choice([f.value for f in UnderlyingCurrency]),
    default=UnderlyingCurrency.ETH.value,
)
@click.option(
    "--columns",
    "-c",
    type=str,
    default=None,
)
@click.pass_context
def fetch_subaccount(ctx, subaccount_id, underlying_currency, columns):
    """Fetch subaccount."""
    print("Fetching subaccount")
    print(f"Subaccount ID: {subaccount_id}")
    print(f"Underlying currency: {underlying_currency}")
    client = ctx.obj["client"]
    subaccount = client.fetch_subaccount(subaccount_id=subaccount_id)
    analyser = PortfolioAnalyser(subaccount)
    print("Positions")
    analyser.print_positions(underlying_currency=underlying_currency, columns=columns)
    print("Total Greeks")
    print(analyser.get_total_greeks(underlying_currency))
    print("Subaccount values")
    print(f"Portfolio Value: ${analyser.get_subaccount_value():.2f}")


@subaccounts.command("create")
@click.pass_context
@click.option(
    "--underlying-currency",
    "-u",
    type=click.Choice([f.value for f in UnderlyingCurrency]),
    default=UnderlyingCurrency.ETH.value,
)
@click.option(
    "--subaccount-type",
    "-s",
    type=click.Choice([f.value for f in SubaccountType]),
    default=SubaccountType.PORTFOLIO.value,
)
@click.option(
    "--collateral-asset",
    "-c",
    type=click.Choice([f.value for f in CollateralAsset]),
    default=CollateralAsset.USDC.value,
)
@click.option(
    "--amount",
    "-a",
    type=float,
    default=0,
)
def create_subaccount(ctx, collateral_asset, underlying_currency, subaccount_type, amount):
    """Create subaccount."""
    underlying_currency = UnderlyingCurrency(underlying_currency)
    subaccount_type = SubaccountType(subaccount_type)
    collateral_asset = CollateralAsset(collateral_asset)
    print(f"Creating subaccount with collateral asset {collateral_asset} and underlying currency {underlying_currency}")
    client = ctx.obj["client"]
    subaccount_id = client.create_subaccount(
        amount=int(amount * 1e6),
        subaccount_type=subaccount_type,
        collateral_asset=collateral_asset,
        underlying_currency=underlying_currency,
    )
    print(subaccount_id)


@orders.command("fetch")
@click.pass_context
@click.option(
    "--instrument-name",
    "-i",
    type=str,
    default=None,
)
@click.option(
    "--label",
    "-l",
    type=str,
    default=None,
)
@click.option(
    "--page",
    "-p",
    type=int,
    default=1,
)
@click.option(
    "--page-size",
    "-s",
    type=int,
    default=100,
)
@click.option(
    "--status",
    "-s",
    type=click.Choice([f.value for f in OrderStatus]),
    default=None,
)
@click.option(
    "--regex",
    "-r",
    type=str,
    default=None,
)
def fetch_orders(ctx, instrument_name, label, page, page_size, status, regex):
    """Fetch orders."""
    print("Fetching orders")
    client = ctx.obj["client"]
    orders = client.fetch_orders(
        instrument_name=instrument_name,
        label=label,
        page=page,
        page_size=page_size,
        status=status,
    )
    # apply the regex if exists to filter the orders
    if regex:
        orders = [o for o in orders if regex in o["instrument_name"]]
    df = pd.DataFrame.from_records(orders)
    print(orders[0])
    instrument_names = df["instrument_name"].unique()
    print(f"Found {len(instrument_names)} instruments")
    print(instrument_names)
    # print the orders
    # perform some analysis
    df['amount'] = pd.to_numeric(df['amount'])
    df['filled_amount'] = pd.to_numeric(df['filled_amount'])
    df['limit_price'] = pd.to_numeric(df['limit_price'])

    buys = df[df['direction'] == 'buy']
    sells = df[df['direction'] == 'sell']
    print("Buys")
    print(buys)
    print("Sells")
    print(sells)

    print("Average buy cost")
    # we determine by the average price of the buys by the amount
    df['cost'] = buys['limit_price'] * buys['amount']
    print(df['cost'].sum())
    amount = buys['amount'].sum()
    print(amount)
    buy_total_cost = df['cost'].sum()
    print(f"Price per unit: {buy_total_cost / amount}")
    print(buy_total_cost / amount)
    print("Average sell cost")
    # we determine by the average price of the buys by the amount
    df['cost'] = sells['limit_price'] * sells['amount']
    print(df['cost'].sum())
    amount = sells['amount'].sum()
    print(amount)
    sell_total_cost = df['cost'].sum()
    print(f"Price per unit: {sell_total_cost / amount}")
    print(sell_total_cost / amount)


@orders.command("cancel")
@click.pass_context
@click.option(
    "--order-id",
    "-o",
    type=str,
)
@click.option(
    "--instrument-name",
    "-i",
    type=str,
)
def cancel_order(ctx, order_id, instrument_name):
    """Cancel order."""
    print("Cancelling order")
    client = ctx.obj["client"]
    result = client.cancel(order_id=order_id, instrument_name=instrument_name)
    print(result)


@orders.command("cancel_all")
@click.pass_context
def cancel_all_orders(ctx):
    """Cancel all orders."""
    print("Cancelling all orders")
    client = ctx.obj["client"]
    result = client.cancel_all()
    print(result)


@orders.command("create")
@click.pass_context
@click.option(
    "--instrument-name",
    "-i",
    type=str,
    required=True,
)
@click.option(
    "--side",
    "-s",
    type=click.Choice(i.value for i in OrderSide),
    required=True,
)
@click.option(
    "--price",
    "-p",
    type=float,
    required=True,
)
@click.option(
    "--amount",
    "-a",
    type=float,
    required=True,
)
@click.option(
    "--order-type",
    "-t",
    type=click.Choice(i.value for i in OrderType),
    default="limit",
)
def create_order(ctx, instrument_name, side, price, amount, order_type):
    """Create order."""
    print("Creating order")
    client = ctx.obj["client"]
    result = client.create_order(
        instrument_name=instrument_name,
        side=OrderSide(side),
        price=price,
        amount=amount,
        order_type=OrderType(order_type),
    )
    print(result)


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter



"""
Class based analyser for portfolios.
"""

from typing import List, Optional

import pandas as pd

pd.set_option('display.precision', 2)


DELTA_COLUMNS = ['delta', 'gamma', 'vega', 'theta']


class PortfolioAnalyser:
    raw_data: List[dict]
    df: pd.DataFrame

    def __init__(self, raw_data: List[dict]):
        self.raw_data = raw_data
        self.positions = pd.DataFrame.from_records(raw_data['positions'])
        self.positions["amount"] = pd.to_numeric(self.positions["amount"])
        for col in DELTA_COLUMNS:
            self.positions[col] = pd.to_numeric(self.positions[col])
            adjusted_greek = self.positions[col] * self.positions.amount
            self.positions[col] = adjusted_greek

        self.positions = self.positions.apply(pd.to_numeric, errors='ignore')

    def get_positions(self, underlying_currency: str) -> pd.DataFrame:
        df = self.positions
        df = df[df['instrument_name'].str.contains(underlying_currency.upper())]
        return df

    def get_open_positions(self, underlying_currency: str) -> pd.DataFrame:
        df = self.get_positions(underlying_currency)
        return df[df['amount'] != 0]

    def get_total_greeks(self, underlying_currency: str) -> pd.DataFrame:
        df = self.get_open_positions(underlying_currency)
        return df[DELTA_COLUMNS].sum()

    def get_subaccount_value(self) -> float:
        return float(self.raw_data['subaccount_value'])

    def print_positions(self, underlying_currency: str, columns: Optional[List[str]] = None):
        df = self.get_open_positions(underlying_currency)
        if columns:
            df = df[[c for c in columns if c not in DELTA_COLUMNS] + DELTA_COLUMNS]
        print(df)



"""
Lyra is a Python library for trading on lyra v2
"""

import pandas as pd
from web3 import Web3

from lyra.base_client import BaseClient
from lyra.http_client import HttpClient

# we set to show 4 decimal places
pd.options.display.float_format = '{:,.4f}'.format


def to_32byte_hex(val):
    return Web3.to_hex(Web3.to_bytes(val).rjust(32, b"\0"))


class LyraClient(BaseClient):
    """Client for the lyra dex."""

    def _create_signature_headers(self):
        """Generate the signature headers."""
        return HttpClient._create_signature_headers(self)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.login_client()



"""
Constants for Lyra.
"""
from lyra.enums import Environment

PUBLIC_HEADERS = {"accept": "application/json", "content-type": "application/json"}

TEST_PRIVATE_KEY = "0xc14f53ee466dd3fc5fa356897ab276acbef4f020486ec253a23b0d1c3f89d4f4"

CONTRACTS = {
    Environment.TEST: {
        "BASE_URL": "https://api-demo.lyra.finance",
        "WS_ADDRESS": "wss://api-demo.lyra.finance/ws",
        "ACTION_TYPEHASH": '0x4d7a9f27c403ff9c0f19bce61d76d82f9aa29f8d6d4b0c5474607d9770d1af17',
        "DOMAIN_SEPARATOR": '0x9bcf4dc06df5d8bf23af818d5716491b995020f377d3b7b64c29ed14e3dd1105',
        "ETH_PERP_ADDRESS": '0x010e26422790C6Cb3872330980FAa7628FD20294',
        "BTC_PERP_ADDRESS": '0xAFB6Bb95cd70D5367e2C39e9dbEb422B9815339D',
        "ETH_OPTION_ADDRESS": '0xBcB494059969DAaB460E0B5d4f5c2366aab79aa1',
        "BTC_OPTION_ADDRESS": '0xAeB81cbe6b19CeEB0dBE0d230CFFE35Bb40a13a7',
        "TRADE_MODULE_ADDRESS": '0x87F2863866D85E3192a35A73b388BD625D83f2be',
        "STANDARD_RISK_MANAGER_ADDRESS": "0x28bE681F7bEa6f465cbcA1D25A2125fe7533391C",
        "BTC_PORTFOLIO_RISK_MANAGER_ADDRESS": "0xbaC0328cd4Af53d52F9266Cdbd5bf46720320A20",
        "ETH_PORTFOLIO_RISK_MANAGER_ADDRESS": "0xDF448056d7bf3f9Ca13d713114e17f1B7470DeBF",
        "CASH_ASSET": "0x6caf294DaC985ff653d5aE75b4FF8E0A66025928",
        "USDC_ASSET": "0xe80F2a02398BBf1ab2C9cc52caD1978159c215BD",
        "DEPOSIT_MODULE_ADDRESS": "0x43223Db33AdA0575D2E100829543f8B04A37a1ec",
        "TRANSFER_MODULE_ADDRESS": "0x0CFC1a4a90741aB242cAfaCD798b409E12e68926",
    },
    Environment.PROD: {
        "BASE_URL": "https://api.lyra.finance",
        "WS_ADDRESS": "wss://api.lyra.finance/ws",
        "ACTION_TYPEHASH": '0x4d7a9f27c403ff9c0f19bce61d76d82f9aa29f8d6d4b0c5474607d9770d1af17',
        "DOMAIN_SEPARATOR": '0xd96e5f90797da7ec8dc4e276260c7f3f87fedf68775fbe1ef116e996fc60441b',
        "ETH_PERP_ADDRESS": '0xAf65752C4643E25C02F693f9D4FE19cF23a095E3',
        "BTC_PERP_ADDRESS": '0xDBa83C0C654DB1cd914FA2710bA743e925B53086',
        "ETH_OPTION_ADDRESS": '0x4BB4C3CDc7562f08e9910A0C7D8bB7e108861eB4',
        "BTC_OPTION_ADDRESS": '0xd0711b9eBE84b778483709CDe62BacFDBAE13623',
        "TRADE_MODULE_ADDRESS": '0xB8D20c2B7a1Ad2EE33Bc50eF10876eD3035b5e7b',
        "STANDARD_RISK_MANAGER_ADDRESS": "0x28c9ddF9A3B29c2E6a561c1BC520954e5A33de5D",
        "BTC_PORTFOLIO_RISK_MANAGER_ADDRESS": "0x45DA02B9cCF384d7DbDD7b2b13e705BADB43Db0D",
        "ETH_PORTFOLIO_RISK_MANAGER_ADDRESS": "0xe7cD9370CdE6C9b5eAbCe8f86d01822d3de205A0",
        "CASH_ASSET": "0x57B03E14d409ADC7fAb6CFc44b5886CAD2D5f02b",
        "USDC_ASSET": "0x6879287835A86F50f784313dBEd5E5cCC5bb8481",
        "DEPOSIT_MODULE_ADDRESS": "0x9B3FE5E5a3bcEa5df4E08c41Ce89C4e3Ff01Ace3",
        "TRANSFER_MODULE_ADDRESS": "0x01259207A40925b794C8ac320456F7F6c8FE2636",
    },
}



"""
Necessary enumarations for the Lyra project.
"""

from enum import Enum


class InstrumentType(Enum):
    """Instrument types."""

    ERC20 = "erc20"
    OPTION = "option"
    PERP = "perp"


class UnderlyingCurrency(Enum):
    """Underlying currencies."""

    ETH = "eth"
    BTC = "btc"


class OrderSide(Enum):
    """Order sides."""

    BUY = "buy"
    SELL = "sell"


class OrderType(Enum):
    """Order types."""

    LIMIT = "limit"
    MARKET = "market"


class OrderStatus(Enum):
    """Order statuses."""

    OPEN = "open"
    FILLED = "filled"
    REJECTED = "rejected"
    CANCELLED = "cancelled"
    EXPIRED = "expired"


class TimeInForce(Enum):
    """Time in force."""

    GTC = "gtc"
    IOC = "ioc"
    FOK = "fok"
    POST_ONLY = "post_only"


class Environment(Enum):
    """Environment."""

    PROD = "prod"
    TEST = "test"


class SubaccountType(Enum):
    """
    Type of sub account
    """

    STANDARD = "standard"
    PORTFOLIO = "portfolio"


class CollateralAsset(Enum):
    """Asset types."""

    USDC = "usdc"


class ActionType(Enum):
    """Action types."""

    DEPOSIT = "deposit"
    TRANSFER = "transfer"


class RfqStatus(Enum):
    """RFQ statuses."""

    OPEN = "open"



"""
Base Client for the lyra dex.
"""
import json
import random
import time
from datetime import datetime

import eth_abi
import requests
from eth_account.messages import encode_defunct
from rich import print
from web3 import Web3
from websocket import WebSocketConnectionClosedException, create_connection

from lyra.constants import CONTRACTS, PUBLIC_HEADERS, TEST_PRIVATE_KEY
from lyra.enums import (
    ActionType,
    CollateralAsset,
    Environment,
    InstrumentType,
    OrderSide,
    OrderStatus,
    OrderType,
    RfqStatus,
    SubaccountType,
    TimeInForce,
    UnderlyingCurrency,
)
from lyra.utils import get_logger


class BaseClient:
    """Client for the lyra dex."""

    def __init__(
        self,
        private_key: str = TEST_PRIVATE_KEY,
        env: Environment = Environment.TEST,
        logger=None,
        verbose=False,
        subaccount_id=None,
        wallet=None,
    ):
        """
        Initialize the LyraClient class.
        """
        self.verbose = verbose
        self.env = env
        self.contracts = CONTRACTS[env]
        self.logger = logger or get_logger()
        self.web3_client = Web3()
        self.signer = self.web3_client.eth.account.from_key(private_key)
        self.wallet = self.signer.address if not wallet else wallet
        print(f"Signing address: {self.signer.address}")
        if wallet:
            print(f"Using wallet: {wallet}")
        if not subaccount_id:
            self.subaccount_id = self.fetch_subaccounts()['subaccount_ids'][0]
        else:
            self.subaccount_id = subaccount_id
        print(f"Using subaccount id: {self.subaccount_id}")

    def sign_authentication_header(self):
        timestamp = str(int(time.time() * 1000))
        msg = encode_defunct(
            text=timestamp,
        )
        signature = self.web3_client.eth.account.sign_message(
            msg, private_key=self.signer._private_key
        ).signature.hex()  # pylint: disable=protected-access
        return {
            'wallet': self.wallet,
            'timestamp': str(timestamp),
            'signature': signature,
        }

    def connect_ws(self):
        ws = create_connection(self.contracts['WS_ADDRESS'], enable_multithread=True, timeout=60)
        return ws

    def create_account(self, wallet):
        """Call the create account endpoint."""
        payload = {"wallet": wallet}
        url = f"{self.contracts['BASE_URL']}/public/create_account"
        result = requests.post(
            headers=PUBLIC_HEADERS,
            url=url,
            json=payload,
        )
        result_code = json.loads(result.content)

        if "error" in result_code:
            raise Exception(result_code["error"])
        return True

    def fetch_instruments(
        self,
        expired=False,
        instrument_type: InstrumentType = InstrumentType.PERP,
        currency: UnderlyingCurrency = UnderlyingCurrency.BTC,
    ):
        """
        Return the tickers.
        First fetch all instrucments
        Then get the ticket for all instruments.
        """
        url = f"{self.contracts['BASE_URL']}/public/get_instruments"
        payload = {
            "expired": expired,
            "instrument_type": instrument_type.value,
            "currency": currency.name,
        }
        response = requests.post(url, json=payload, headers=PUBLIC_HEADERS)
        results = response.json()["result"]
        return results

    def fetch_subaccounts(self):
        """
        Returns the subaccounts for a given wallet
        """
        url = f"{self.contracts['BASE_URL']}/private/get_subaccounts"
        payload = {"wallet": self.wallet}
        headers = self._create_signature_headers()
        response = requests.post(url, json=payload, headers=headers)
        results = json.loads(response.content)["result"]
        return results

    def fetch_subaccount(self, subaccount_id):
        """
        Returns information for a given subaccount
        """
        url = f"{self.contracts['BASE_URL']}/private/get_subaccount"
        payload = {"subaccount_id": subaccount_id}
        headers = self._create_signature_headers()
        response = requests.post(url, json=payload, headers=headers)
        results = response.json()["result"]
        return results

    def create_order(
        self,
        price,
        amount,
        instrument_name: str,
        reduce_only=False,
        side: OrderSide = OrderSide.BUY,
        order_type: OrderType = OrderType.LIMIT,
        time_in_force: TimeInForce = TimeInForce.GTC,
    ):
        """
        Create the order.
        """
        if side.name.upper() not in OrderSide.__members__:
            raise Exception(f"Invalid side {side}")
        order = self._define_order(
            instrument_name=instrument_name,
            price=price,
            amount=amount,
            side=side,
        )
        _currency = UnderlyingCurrency[instrument_name.split("-")[0]]
        if instrument_name.split("-")[1] == "PERP":
            instruments = self.fetch_instruments(instrument_type=InstrumentType.PERP, currency=_currency)
            instruments = {i['instrument_name']: i for i in instruments}
            base_asset_sub_id = instruments[instrument_name]['base_asset_sub_id']
            instrument_type = InstrumentType.PERP
        else:
            instruments = self.fetch_instruments(instrument_type=InstrumentType.OPTION, currency=_currency)
            instruments = {i['instrument_name']: i for i in instruments}
            base_asset_sub_id = instruments[instrument_name]['base_asset_sub_id']
            instrument_type = InstrumentType.OPTION

        signed_order = self._sign_order(order, base_asset_sub_id, instrument_type, _currency)
        response = self.submit_order(signed_order)
        return response

    def _define_order(
        self,
        instrument_name: str,
        price: float,
        amount: float,
        side: OrderSide,
        time_in_force: TimeInForce = TimeInForce.GTC,
    ):
        """
        Define the order, in preparation for encoding and signing
        """
        ts = int(datetime.now().timestamp() * 1000)
        return {
            'instrument_name': instrument_name,
            'subaccount_id': self.subaccount_id,
            'direction': side.name.lower(),
            'limit_price': price,
            'amount': amount,
            'signature_expiry_sec': int(ts) + 3000,
            'max_fee': '200.01',
            'nonce': int(f"{int(ts)}{random.randint(100, 999)}"),
            'signer': self.signer.address,
            'order_type': 'limit',
            'mmp': False,
            'time_in_force': time_in_force.value,
            'signature': 'filled_in_below',
        }

    def submit_order(self, order):
        id = str(int(time.time()))
        self.ws.send(json.dumps({'method': 'private/order', 'params': order, 'id': id}))
        while True:
            message = json.loads(self.ws.recv())
            if message['id'] == id:
                try:
                    return message['result']['order']
                except KeyError as error:
                    print(message)
                    raise Exception(f"Unable to submit order {message}") from error

    def _encode_trade_data(self, order, base_asset_sub_id, instrument_type, currency):
        encoded_data = eth_abi.encode(
            ['address', 'uint256', 'int256', 'int256', 'uint256', 'uint256', 'bool'],
            [
                self.contracts[f'{currency.name}_{instrument_type.name}_ADDRESS'],
                int(base_asset_sub_id),
                self.web3_client.to_wei(order['limit_price'], 'ether'),
                self.web3_client.to_wei(order['amount'], 'ether'),
                self.web3_client.to_wei(order['max_fee'], 'ether'),
                order['subaccount_id'],
                order['direction'] == 'buy',
            ],
        )

        return self.web3_client.keccak(encoded_data)

    def _sign_order(self, order, base_asset_sub_id, instrument_type, currency):
        trade_module_data = self._encode_trade_data(order, base_asset_sub_id, instrument_type, currency)
        encoded_action_hash = eth_abi.encode(
            ['bytes32', 'uint256', 'uint256', 'address', 'bytes32', 'uint256', 'address', 'address'],
            [
                bytes.fromhex(self.contracts['ACTION_TYPEHASH'][2:]),
                order['subaccount_id'],
                order['nonce'],
                self.contracts['TRADE_MODULE_ADDRESS'],
                trade_module_data,
                order['signature_expiry_sec'],
                self.wallet,
                order['signer'],
            ],
        )

        action_hash = self.web3_client.keccak(encoded_action_hash)
        encoded_typed_data_hash = "".join(['0x1901', self.contracts['DOMAIN_SEPARATOR'][2:], action_hash.hex()[2:]])
        typed_data_hash = self.web3_client.keccak(hexstr=encoded_typed_data_hash)
        order['signature'] = self.signer.signHash(typed_data_hash).signature.hex()
        return order

    def _sign_quote(self, quote):
        """
        Sign the quote
        """
        rfq_module_data = self._encode_quote_data(quote)
        return self._sign_quote_data(quote, rfq_module_data)

    def _encode_quote_data(self, quote, underlying_currency: UnderlyingCurrency = UnderlyingCurrency.ETH):
        """
        Convert the quote to encoded data.
        """
        instruments = self.fetch_instruments(instrument_type=InstrumentType.OPTION, currency=underlying_currency)
        ledgs_to_subids = {i['instrument_name']: i['base_asset_sub_id'] for i in instruments}
        dir_sign = 1 if quote['direction'] == 'buy' else -1
        quote['price'] = '10'

        def encode_leg(leg):
            print(quote)
            sub_id = ledgs_to_subids[leg['instrument_name']]
            leg_sign = 1 if leg['direction'] == 'buy' else -1
            signed_amount = self.web3_client.to_wei(leg['amount'], 'ether') * leg_sign * dir_sign
            return [
                self.contracts[f"{underlying_currency.name}_OPTION_ADDRESS"],
                sub_id,
                self.web3_client.to_wei(quote['price'], 'ether'),
                signed_amount,
            ]

        encoded_legs = [encode_leg(leg) for leg in quote['legs']]
        rfq_data = [self.web3_client.to_wei(quote['max_fee'], 'ether'), encoded_legs]

        encoded_data = eth_abi.encode(
            # ['uint256(address,uint256,uint256,int256)[]'],
            [
                'uint256',
                'address',
                'uint256',
                'int256',
            ],
            [rfq_data],
        )
        return self.web3_client.keccak(encoded_data)

    @property
    def ws(self):
        if not hasattr(self, '_ws'):
            self._ws = self.connect_ws()
        if not self._ws.connected:
            self._ws = self.connect_ws()
        return self._ws

    def login_client(
        self,
        retries=3,
    ):
        login_request = {
            'method': 'public/login',
            'params': self.sign_authentication_header(),
            'id': str(int(time.time())),
        }
        try:
            self.ws.send(json.dumps(login_request))
            # we need to wait for the response
            while True:
                message = json.loads(self.ws.recv())
                if message['id'] == login_request['id']:
                    if "result" not in message:
                        raise Exception(f"Unable to login {message}")
                    break
        except (WebSocketConnectionClosedException, Exception) as error:
            if retries:
                time.sleep(1)
                self.login_client(retries=retries - 1)
            raise error

    def fetch_ticker(self, instrument_name):
        """
        Fetch the ticker for a given instrument name.
        """
        url = f"{self.contracts['BASE_URL']}/public/get_ticker"
        payload = {"instrument_name": instrument_name}
        response = requests.post(url, json=payload, headers=PUBLIC_HEADERS)
        results = json.loads(response.content)["result"]
        return results

    def fetch_orders(
        self,
        instrument_name: str = None,
        label: str = None,
        page: int = 1,
        page_size: int = 100,
        status: OrderStatus = None,
    ):
        """
        Fetch the orders for a given instrument name.
        """
        url = f"{self.contracts['BASE_URL']}/private/get_orders"
        payload = {"instrument_name": instrument_name, "subaccount_id": self.subaccount_id}
        for key, value in {"label": label, "page": page, "page_size": page_size, "status": status}.items():
            if value:
                payload[key] = value
        headers = self._create_signature_headers()
        response = requests.post(url, json=payload, headers=headers)
        results = response.json()["result"]['orders']
        return results

    def cancel(self, order_id, instrument_name):
        """
        Cancel an order
        """

        id = str(int(time.time()))
        payload = {"order_id": order_id, "subaccount_id": self.subaccount_id, "instrument_name": instrument_name}
        self.ws.send(json.dumps({'method': 'private/cancel', 'params': payload, 'id': id}))
        while True:
            message = json.loads(self.ws.recv())
            if message['id'] == id:
                return message['result']

    def cancel_all(self):
        """
        Cancel all orders
        """
        id = str(int(time.time()))
        payload = {"subaccount_id": self.subaccount_id}
        self.login_client()
        self.ws.send(json.dumps({'method': 'private/cancel_all', 'params': payload, 'id': id}))
        while True:
            message = json.loads(self.ws.recv())
            if message['id'] == id:
                return message['result']

    def get_positions(self):
        """
        Get positions
        """
        url = f"{self.contracts['BASE_URL']}/private/get_positions"
        payload = {"subaccount_id": self.subaccount_id}
        headers = self._create_signature_headers()
        response = requests.post(url, json=payload, headers=headers)
        results = response.json()["result"]['positions']
        return results

    def get_collaterals(self):
        """
        Get collaterals
        """
        url = f"{self.contracts['BASE_URL']}/private/get_collaterals"
        payload = {"subaccount_id": self.subaccount_id}
        headers = self._create_signature_headers()
        response = requests.post(url, json=payload, headers=headers)
        results = response.json()["result"]['collaterals']
        return results.pop()

    def fetch_tickers(
        self,
        instrument_type: InstrumentType = InstrumentType.OPTION,
        currency: UnderlyingCurrency = UnderlyingCurrency.BTC,
    ):
        """
        Fetch tickers using the ws connection
        """
        instruments = self.fetch_instruments(instrument_type=instrument_type, currency=currency)
        instrument_names = [i['instrument_name'] for i in instruments]
        id_base = str(int(time.time()))
        ids_to_instrument_names = {
            f'{id_base}_{enumerate}': instrument_name for enumerate, instrument_name in enumerate(instrument_names)
        }
        for id, instrument_name in ids_to_instrument_names.items():
            payload = {"instrument_name": instrument_name}
            self.ws.send(json.dumps({'method': 'public/get_ticker', 'params': payload, 'id': id}))
            time.sleep(0.05)  # otherwise we get rate limited...
        results = {}
        while ids_to_instrument_names:
            message = json.loads(self.ws.recv())
            if message['id'] in ids_to_instrument_names:
                results[message['result']['instrument_name']] = message['result']
                del ids_to_instrument_names[message['id']]
        return results

    def create_subaccount(
        self,
        amount=0,
        subaccount_type: SubaccountType = SubaccountType.STANDARD,
        collateral_asset: CollateralAsset = CollateralAsset.USDC,
        underlying_currency: UnderlyingCurrency = UnderlyingCurrency.ETH,
    ):
        """
        Create a subaccount.
        """
        url = f"{self.contracts['BASE_URL']}/private/create_subaccount"
        _, nonce, expiration = self.get_nonce_and_signature_expiry()
        if subaccount_type is SubaccountType.STANDARD:
            contract_key = f"{subaccount_type.name}_RISK_MANAGER_ADDRESS"
        elif subaccount_type is SubaccountType.PORTFOLIO:
            if not collateral_asset:
                raise Exception("Underlying currency must be provided for portfolio subaccounts")
            contract_key = f"{underlying_currency.name}_{subaccount_type.name}_RISK_MANAGER_ADDRESS"
        else:
            raise Exception(f"Invalid subaccount type {subaccount_type}")
        payload = {
            "amount": f"{amount}",
            "asset_name": collateral_asset.name,
            "margin_type": "SM" if subaccount_type is SubaccountType.STANDARD else "PM",
            'nonce': nonce,
            "signature": "string",
            "signature_expiry_sec": expiration,
            "signer": self.signer.address,
            "wallet": self.wallet,
        }
        if subaccount_type is SubaccountType.PORTFOLIO:
            payload['currency'] = underlying_currency.name
        encoded_deposit_data = self._encode_deposit_data(
            amount=amount,
            contract_key=contract_key,
        )
        action_hash = self._generate_action_hash(
            subaccount_id=0,  # as we are depositing to a new subaccount.
            nonce=nonce,
            expiration=expiration,
            encoded_deposit_data=encoded_deposit_data,
        )

        typed_data_hash = self._generate_typed_data_hash(
            action_hash=action_hash,
        )

        signature = self.signer.signHash(typed_data_hash).signature.hex()
        payload['signature'] = signature
        print(f"Payload: {payload}")

        headers = self._create_signature_headers()
        response = requests.post(url, json=payload, headers=headers)

        if "error" in response.json():
            raise Exception(response.json()["error"])
        print(response.text)
        if "result" not in response.json():
            raise Exception(f"Unable to create subaccount {response.json()}")
        return response.json()["result"]

    def _encode_deposit_data(self, amount: int, contract_key: str):
        """Encode the deposit data"""

        encoded_data = eth_abi.encode(
            ['uint256', 'address', 'address'],
            [
                int(amount * 1e6),
                self.contracts["CASH_ASSET"],
                self.contracts[contract_key],
            ],
        )
        print(f"Encoded data: {encoded_data}")
        return self.web3_client.keccak(encoded_data)

    def get_nonce_and_signature_expiry(self):
        """
        Returns the nonce and signature expiry
        """
        ts = int(datetime.now().timestamp() * 1000)
        nonce = int(f"{int(ts)}{random.randint(100, 999)}")
        expiration = int(ts) + 6000
        return ts, nonce, expiration

    def _generate_typed_data_hash(
        self,
        action_hash: bytes,
    ):
        """Generate the typed data hash."""

        encoded_typed_data_hash = "".join(['0x1901', self.contracts['DOMAIN_SEPARATOR'][2:], action_hash.hex()[2:]])
        typed_data_hash = self.web3_client.keccak(hexstr=encoded_typed_data_hash)
        return typed_data_hash

    def transfer_collateral(self, amount: int, to: str, asset: CollateralAsset):
        """
        Transfer collateral
        """

        ts = int(datetime.now().timestamp() * 1000)
        nonce = int(f"{int(ts)}{random.randint(100, 499)}")
        nonce_2 = int(f"{int(ts)}{random.randint(500, 999)}")
        expiration = int(datetime.now().timestamp() + 10000)

        url = f"{self.contracts['BASE_URL']}/private/transfer_erc20"
        _, nonce, expiration = self.get_nonce_and_signature_expiry()
        transfer = {
            "address": self.contracts["CASH_ASSET"],
            "amount": int(amount),
            "sub_id": 0,
        }
        print(f"Transfering to {to} amount {amount} asset {asset.name}")

        encoded_data = self.encode_transfer(
            amount=amount,
            to=to,
        )

        action_hash_1 = self._generate_action_hash(
            subaccount_id=self.subaccount_id,
            nonce=nonce,
            expiration=expiration,
            encoded_deposit_data=encoded_data,
            action_type=ActionType.TRANSFER,
        )

        from_signed_action_hash = self._generate_signed_action(
            action_hash=action_hash_1,
            nonce=nonce,
            expiration=expiration,
        )

        print(f"from_signed_action_hash: {from_signed_action_hash}")
        print(f"From action hash: {action_hash_1.hex()}")

        action_hash_2 = self._generate_action_hash(
            subaccount_id=to,
            nonce=nonce_2,
            expiration=expiration,
            encoded_deposit_data=self.web3_client.keccak(bytes.fromhex('')),
            action_type=ActionType.TRANSFER,
        )
        to_signed_action_hash = self._generate_signed_action(
            action_hash=action_hash_2,
            nonce=nonce_2,
            expiration=expiration,
        )

        print(f"To action hash: {action_hash_2.hex()}")
        print(f"To signed action hash: {to_signed_action_hash}")
        payload = {
            "subaccount_id": self.subaccount_id,
            "recipient_subaccount_id": to,
            "sender_details": {
                "nonce": nonce,
                "signature": "string",
                "signature_expiry_sec": expiration,
                "signer": self.signer.address,
            },
            "recipient_details": {
                "nonce": nonce_2,
                "signature": "string",
                "signature_expiry_sec": expiration,
                "signer": self.signer.address,
            },
            "transfer": transfer,
        }
        payload['sender_details']['signature'] = from_signed_action_hash['signature']
        payload['recipient_details']['signature'] = to_signed_action_hash['signature']

        print(payload)
        headers = self._create_signature_headers()
        response = requests.post(url, json=payload, headers=headers)

        print(response.json())

        if "error" in response.json():
            raise Exception(response.json()["error"])
        if "result" not in response.json():
            raise Exception(f"Unable to transfer collateral {response.json()}")
        return response.json()["result"]

    def encode_transfer(self, amount: int, to: str, asset_sub_id=0, signature_expiry=300):
        """
        Encode the transfer
        const encoder = ethers.AbiCoder.defaultAbiCoder();
        const TransferDataABI = ['(uint256,address,(address,uint256,int256)[])'];
        const signature_expiry = getUTCEpochSec() + 300;

        const fromTransfers = [
          [
            assetAddress,
            assetSubId,
            ethers.parseUnits(amount, 18), // Amount in wei
          ],
        ];

        const fromTransferData = [
          toAccount.subaccountId,
          "0x0000000000000000000000000000000000000000", // manager (if new account)`
          fromTransfers,
        ];

        const fromEncodedData = encoder.encode(TransferDataABI, [fromTransferData]);
        """
        transfer_data_abi = ["(uint256,address,(address,uint256,int256)[])"]

        from_transfers = [
            [
                self.contracts["CASH_ASSET"],
                asset_sub_id,
                self.web3_client.to_wei(amount, 'ether'),
            ]
        ]

        from_transfer_data = [
            int(to),
            "0x0000000000000000000000000000000000000000",
            from_transfers,
        ]

        from_encoded_data = eth_abi.encode(transfer_data_abi, [from_transfer_data])
        print(f"From transfers: {from_transfers}")
        print(f"From transfer data: {from_transfer_data}")
        print(f"From encoded data: {from_encoded_data.hex()}")

        # need to add the signature expiry
        return self.web3_client.keccak(from_encoded_data)

    def _generate_action_hash(
        self,
        subaccount_id: int,
        nonce: int,
        expiration: int,
        encoded_deposit_data: bytes,
        action_type: ActionType = ActionType.DEPOSIT,
    ):
        """Handle the deposit to a new subaccount."""
        encoded_action_hash = eth_abi.encode(
            ['bytes32', 'uint256', 'uint256', 'address', 'bytes32', 'uint256', 'address', 'address'],
            [
                bytes.fromhex(self.contracts['ACTION_TYPEHASH'][2:]),
                subaccount_id,
                nonce,
                self.contracts[f'{action_type.name}_MODULE_ADDRESS'],
                encoded_deposit_data,
                expiration,
                self.wallet,
                self.signer.address,
            ],
        )
        return self.web3_client.keccak(encoded_action_hash)

    def _generate_signed_action(self, action_hash: bytes, nonce: int, expiration: int):
        """Generate the signed action."""
        encoded_typed_data_hash = "".join(['0x1901', self.contracts['DOMAIN_SEPARATOR'][2:], action_hash.hex()[2:]])
        typed_data_hash = self.web3_client.keccak(hexstr=encoded_typed_data_hash)
        signature = self.signer.signHash(typed_data_hash).signature.hex()
        return {
            "nonce": nonce,
            "signature": signature,
            "signature_expiry_sec": expiration,
            "signer": self.signer.address,
        }

    def get_mmp_config(self, subaccount_id: int, currency: UnderlyingCurrency = None):
        """Get the mmp config."""
        url = f"{self.contracts['BASE_URL']}/private/get_mmp_config"
        payload = {"subaccount_id": self.subaccount_id}
        if currency:
            payload['currency'] = currency.name
        headers = self._create_signature_headers()
        response = requests.post(url, json=payload, headers=headers)
        results = response.json()["result"]
        return results

    def set_mmp_config(
        self,
        subaccount_id,
        currency: UnderlyingCurrency,
        mmp_frozen_time: int,
        mmp_interval: int,
        mmp_amount_limit: str,
        mmp_delta_limit: str,
    ):
        """Set the mmp config."""
        url = f"{self.contracts['BASE_URL']}/private/set_mmp_config"
        payload = {
            "subaccount_id": subaccount_id,
            "currency": currency.name,
            "mmp_frozen_time": mmp_frozen_time,
            "mmp_interval": mmp_interval,
            "mmp_amount_limit": mmp_amount_limit,
            "mmp_delta_limit": mmp_delta_limit,
        }
        headers = self._create_signature_headers()
        response = requests.post(url, json=payload, headers=headers)
        results = response.json()["result"]
        return results

    def send_rfq(self, rfq):
        """Send an RFQ."""
        url = f"{self.contracts['BASE_URL']}/private/send_rfq"
        headers = self._create_signature_headers()
        response = requests.post(url, json=rfq, headers=headers)
        results = response.json()["result"]
        return results

    def poll_rfqs(self):
        """
        Poll RFQs.
            type RfqResponse = {
              subaccount_id: number,
              creation_timestamp: number,
              last_update_timestamp: number,
              status: string,
              cancel_reason: string,
              rfq_id: string,
              valid_until: number,
              legs: Array<RfqLeg>
            }
        """
        url = f"{self.contracts['BASE_URL']}/private/poll_rfqs"
        headers = self._create_signature_headers()
        params = {
            "subaccount_id": self.subaccount_id,
            "status": RfqStatus.OPEN.value,
        }
        response = requests.post(url, headers=headers, params=params)
        results = response.json()["result"]
        return results

    def send_quote(self, quote):
        """Send a quote."""
        url = f"{self.contracts['BASE_URL']}/private/send_quote"
        headers = self._create_signature_headers()
        response = requests.post(url, json=quote, headers=headers)
        results = response.json()["result"]
        return results

    #   pricedLegs[0].price = direction == 'buy' ? '160' : '180';
    #   pricedLegs[1].price = direction == 'buy' ? '70' : '50';
    #   return {
    #     subaccount_id: subaccount_id_maker,
    #     rfq_id: rfq_response.rfq_id,
    #     legs: pricedLegs,
    #     direction: direction,
    #     max_fee: '10',
    #     nonce: Number(`${Date.now()}${Math.round(Math.random() * 999)}`),
    #     signer: wallet.address,
    #     signature_expiry_sec: Math.floor(Date.now() / 1000 + 350),
    #     signature: "filled_in_below"
    #   };
    # }

    def create_quote_object(
        self,
        rfq_id,
        legs,
        direction,
    ):
        """Create a quote object."""
        _, nonce, expiration = self.get_nonce_and_signature_expiry()
        return {
            "subaccount_id": self.subaccount_id,
            "rfq_id": rfq_id,
            "legs": legs,
            "direction": direction,
            "max_fee": '10.0',
            "nonce": nonce,
            "signer": self.signer.address,
            "signature_expiry_sec": expiration,
            "signature": "filled_in_below",
        }



"""
Class to handle base websocket client
"""

import time

from eth_account.messages import encode_defunct
from web3 import Web3

from lyra.base_client import BaseClient


class WsClient(BaseClient):
    def _create_signature_headers(self):
        """
        Create the signature headers
        """
        timestamp = str(int(time.time() * 1000))
        msg = encode_defunct(
            text=timestamp,
        )
        signature = self.signer.sign_message(msg)
        return {
            "X-LyraWallet": self.wallet,
            "X-LyraTimestamp": timestamp,
            "X-LyraSignature": Web3.to_hex(signature.signature),
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ws = self.connect_ws()
        self.login_client()
